(* parser_utils.ml - This utility is used for testing purposes.     *)
(* It outputs the results of an AST in string-readable form          *)

open Type
open Ast

let string_of_taty_var = function
    Int -> "TYPE_INT"
  | Float -> "TYPE_FLOAT"
  | Char -> "TYPE_CHAR"
  | String -> "TYPE_STRING"
  | Boolean -> "TYPE_BOOL"
  | Table -> "TYPE_TABLE"
  | Line -> "TYPE_LINE"
  | Cell -> "TYPE_CELL"

let string_of_taty_ret = function
    RInt -> "TYPE_INT"
  | RFloat -> "TYPE_FLOAT"
  | RChar -> "TYPE_CHAR"
  | RString -> "TYPE_STRING"
  | RBoolean -> "TYPE_BOOL"
  | RTable -> "TYPE_TABLE"
  | RLine -> "TYPE_LINE"
  | RCell -> "TYPE_CELL"
  | RVoid -> "TYPE_VOID"

let string_of_literal = function
    IntL(l) -> "LITERAL_INT:[" ^ (string_of_int l) ^ "]"
  | FloatL(l) -> "LITERAL_FLOAT:[" ^ (string_of_float l) ^ "]"
  | CharL(l) -> "LITERAL_CHAR:[" ^ (Char.escaped l) ^ "]"
  | BoolL(l) -> "LITERAL_BOOL:[" ^ (if l then "#TRUE#" else "#FALSE#") ^ "]"
  | StringL(l) -> "LITERAL_STRING:[" ^ l ^ "]"

let string_of_binop = function
   Add -> "ADD"
 | Sub -> "SUB"
 | Mult -> "MULT"
 | Div -> "DIV"
 | Equal -> "EQUAL"
 | Neq -> "NEQ"
 | Less_than -> "LT"
 | Leq -> "LTE"
 | Greater_than -> "GT"
 | Geq -> "GTE"
 | Or -> "OR"
 | And -> "AND"
 | Mod ->"MOD"

let string_of_uniop = function
   Not -> "NOT"

let rec string_of_expr = function
    Literal(l) -> "EXPR_LITERAL:[" ^ string_of_literal l ^ "]"
  | TableAsTable(s,e1,e2,e3,e4) -> "EXPR_TABLE_AS_TABLE:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "," ^ (string_of_expr e3) ^ "," ^ (string_of_expr e4) ^ "]"
  | TableAsLine_L(s,e1,e2,e3) -> "EXPR_TABLE_AS_LINE_L:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "," ^ (string_of_expr e3) ^ "]"
  | TableAsLine_R(s,e1,e2,e3) -> "EXPR_TABLE_AS_LINE_R:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "," ^ (string_of_expr e3) ^ "]"
  | TableAsCell(s,e1,e2) -> "EXPR_TABLE_AS_CELL:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "]"
  | TableAsCellValue(s,e1,e2) -> "EXPR_TABLE_AS_CELL_Value:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "]"
  | LineAsLine(s,e1,e2) -> "EXPR_LINE_AS_LINE:[" ^ s ^ "," ^ (string_of_expr e1) ^ "," ^ (string_of_expr e2) ^ "]"
  | LineAsCell(s,e1) -> "EXPR_LINE_AS_CELL:[" ^ s ^ "," ^ (string_of_expr e1) ^ "]"
  | LineAsCellValue(s,e1) -> "EXPR_LINE_AS_CELL_Value:[" ^ s ^ "," ^ (string_of_expr e1) ^ "]"
  | CellAsCell(s) -> "EXPR_CELL_AS_CELL:[" ^s^"]"
  | NewTotalCell(s) -> "EXPR_New_Total_CELL:[" ^s^"]"
  | Id(s) -> "EXPR_ID:[" ^ s ^ "]"
  | Binop(e1, o, e2) -> "EXPR_BINOP:[" ^ string_of_expr e1 ^ "," ^ string_of_binop o ^ "," ^ string_of_expr e2 ^ "]"
  | Assign(e1, e2) -> "EXPR_ASSIGN:[" ^ string_of_expr e1 ^ "," ^ string_of_expr e2 ^ "]"
  | Call(f, el) ->
      "EXPR_CALL:[" ^ f ^ ",LIST:[" ^ String.concat ", " (List.map string_of_expr el) ^ "]]"
  | Uniop(o, expr) -> "EXPR_UNIOP:[" ^ string_of_uniop o ^ "," ^ string_of_expr expr ^ "]"
  | All -> "ALL"
  | Null -> "NULL"
  | UnaryMinus(e) -> "EXPR_UNARYMINUS:[" ^ string_of_expr e ^ "]"
let string_of_init = function
    NullInit(t, s) -> "INIT_NULL:[" ^ string_of_taty_var t ^ "," ^ s ^ "]"
  | VarInit(t,s,e) -> "INIT_VAR:[" ^ string_of_taty_var t ^ "," ^ s ^ "," ^ (string_of_expr e) ^ "]"
  | TableInit(t1,s,e1,e2,t2) -> "INIT_TABLE:[" ^ string_of_taty_var t1 ^ "," ^ s ^ "," ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_taty_var t2 ^ "]"

let rec string_of_stmt = function
    Block(stmts) ->
      "STMT_BLOCK:[LIST:[" ^ String.concat "," (List.map string_of_stmt stmts) ^ "]]"
  | Expr(expr) -> "STMT_EXPR:[" ^ string_of_expr expr ^ "]";
  | Return(expr) -> "STMT_RETURN:[" ^ string_of_expr expr ^ "]";
  | Print(expr) -> "STMT_PRINT:[" ^ string_of_expr expr ^ "]";
		
	| If(e, s, NoElseMark) -> "STMT_IF_NOELSE:[" ^ string_of_expr e ^ "," ^ string_of_stmt s ^ "]"
  | If(e, s1, s2) ->  "STMT_IF:[" ^ string_of_expr e ^ "," ^
      string_of_stmt s1 ^ "," ^ string_of_stmt s2 ^ "]"
  | For(e1, e2, e3, s) ->
      "STMT_FOR:[" ^ string_of_expr e1  ^ "," ^ string_of_expr e2 ^ "," ^
      string_of_expr e3  ^ "," ^ string_of_stmt s ^ "]"
  | While(e, s) -> "STMT_WHILE:[" ^ string_of_expr e ^ "," ^ string_of_stmt s ^ "]"
  | NoElseMark -> "STMT_NOELSEMARK"
  | VarDecl(s)->string_of_init s
  | Break -> "STMT_BREAK"


let string_of_funcSig = function
  FuncSig(t, s) -> "FUNCSIG:[" ^ string_of_taty_ret t ^ "," ^ s ^ "]"

let string_of_varSig = function
  VarSig(t, s) -> "VARSIG:[" ^ string_of_taty_var t ^ "," ^ s ^ "]"

let string_of_func_decl fdecl =
  "FUNCDECL:[" ^ string_of_funcSig fdecl.fname ^ ",LIST:[" ^
  String.concat "," (List.map string_of_varSig fdecl.formals) ^ "],LIST:[" ^
  string_of_stmt fdecl.body ^ "]]"

let string_of_construct = function
    GlobalVar(i) -> "CONSTRUCT_GLOBALVAR:[" ^ (string_of_init i) ^ "]"
  | FuncDef(f) -> "CONSTRUCT_FUNCDEF:[" ^ (string_of_func_decl f) ^ "]"

let string_of_program construct_list =
  "PROGRAM:[LIST:[" ^ (String.concat "," (List.map string_of_construct construct_list)) ^ "]]\n"
