(* translate.ml - performs semantic checking on AST 								*)

(* Note on var_scope vs. und_scope: Due to the fact that in TaML 				*)
(* tables/lines/cells can hold int OR float, we made a differentiation	*)
(* between general "variable type" and "underlying type." For 					*)
(* table/line/cell the variable type is table/line/cell respectively 		*)
(* while the underlying type will be int or float -- depending on what 	*)
(* the object holds.  For all other types -- int/float/string/char/bool *)
(* variable type = underlying type *)


open Type
open Ast
open Sast

type mode = Quiet | Verbose (* compiler directives *)

let print_warning mode w = match mode with 
  | Verbose -> print_string("/* WARNING: " ^ w ^ "*/\n")
  | Quiet   -> print_string ""


(* Exception types *) (* TODO - use these instead of generic error everywhere *)
exception ReturnException of string (* Missing return statement in function *)
exception VariableNotFoundException of string (* Using an undeclared variable *)
exception FunctionNotFoundException of string (* Using an undeclared function *)
exception BinopException of string (* Mismatching types in binop *)
exception OtherException of string (* Misc exception *)
exception AssignmentException of string (* Assigning incompatable things *)
exception FunctionException of string (* Incorrect call to built in functions *)
exception WhileExprException of string (* Incorrect specification of while *)
exception IfExprException of string (* Incorrect specification of if *)
exception ReturnTypeException of string (* Mismatching return statement *)
exception MainNotFoundException of string (* Missing main function *)
exception Error of string (* Placeholder - TODO remove *)


(* HELPER FUNCTION: convert normal types to return types*)
(* For simplicity we "used return type" everywhere in semantic checking *)
let var_taty_to_ret_taty t = match t with
    Int     -> RInt
  | Float   -> RFloat
  | Char    -> RChar
  | String  -> RString
  | Boolean -> RBoolean
   (*TaML exclusive*)
  | Table   -> RTable
  | Line    -> RLine
  | Cell    -> RCell

(* HELPER FUNCTION: return types to string representation*)
let taty_ret_to_string tp = match tp with
    RInt 			-> "RInt"
  | RFloat 		-> "RFloat"
  | RChar 		-> "RChar"
  | RString  	-> "RString"
  | RBoolean 	-> "RBoolean"
   (*TaML exclusive*)
  | RTable 		-> "RTable"
  | RLine 		-> "RLine"
  | RCell 		-> "RCell"
  | RVoid 		-> "RVoid"


type symbol_table = {
  parent : symbol_table option;
  variables : (string * taty_ret) list
}


(* all function names are global -- no nested funcs allowed, so  *)
(* just build a table of all functions upon program initialization*)
type function_table = {
  functions: (string * taty_ret * taty_var list * stmt) list
}


type translation_environment = {
  return_type: taty_ret; 		 (* Function's return type *)
  return_seen: bool; 				 (* Have we seen a return statement *)
  var_scope: symbol_table;   (* symbol table for vars *)
  und_scope: symbol_table;   (* symbol table for underlying types *)
  fun_scope: function_table; (* list of functions *)
  break_possible: bool; 		 (* can we break out of this *)
}


(* HEPLER FUNCTION: Add variable to an envirionment's var_scope*)
let add_to_var_table env s t =
  (* let () = print_string ("Adding to var_scope - name: " ^ s ^ " type: " ^ taty_ret_to_string t ^ "\n\n") in *)
  let new_vars = (s,t) :: env.var_scope.variables in
  let new_sym_table = {parent = env.var_scope.parent; variables = new_vars} in
  let new_env = {env with var_scope = new_sym_table} in
  new_env


(* HEPLER FUNCTION: Add variable to an envirionment's und_scope*)
let add_to_und_table env s t =
  (* let () = print_string ("Adding to und_scope - name: " ^ s ^ " type: " ^ taty_ret_to_string t ^ "\n\n") in *)
  let new_vars = (s,t) :: env.und_scope.variables in
  let new_sym_table = {parent = env.und_scope.parent; variables = new_vars} in
  let new_env = {env with und_scope = new_sym_table} in
  new_env


(* HEPLER FUNCTION: Search for variable in the symbol tables *)
let rec find_var_table (var_scope : symbol_table) name =
  try
    List.find (fun (s,_) -> s = name) var_scope.variables
  with Not_found ->
    match var_scope.parent with
        Some(parent) -> find_var_table parent name
      | _ -> raise Not_found


(* HEPLER FUNCTION: Search for variable in the symbol tables in "underlying type" *)
let rec find_und_table (und_scope : symbol_table) name =
  try
    (* let () = List.iter (fun (s,_) -> print_string ("Checked: "^s)) und_scope.variables in *)
    List.find (fun (s,_) -> s = name) und_scope.variables
  with Not_found ->
    match und_scope.parent with
        Some(parent) -> find_und_table parent name
      | _ -> raise Not_found


(* HEPLER FUNCTION: Search for function in our environment *)
let rec find_function (fun_scope : function_table) name =
    List.find (fun (s,_,_,_) -> s = name) fun_scope.functions

let get_better_und_type t1 t2 = match (t1,t2) with
  (RCell, _) as a -> snd a
  | (_, RCell) as a -> fst a
  | (_,_) as a -> fst a

(* HEPLER FUNCTION: Represent all table/line/cell return types as 	*)
(* "cell return type" for simplicity.  Whether these hold int/float *)
(* can only be dynamically determined at run time so this can't   	*)
(* be semantically checked at the moment *)
let default_und_type tp = match tp with
    RTable 	-> RCell
  | RLine 	-> RCell
  | RCell 	-> RCell
  | _ as tp -> tp


(* HEPLER FUNCTION *)
let get_type_for_literal l = match l with
    IntL(l) 		-> Int
  | FloatL(l) 	-> Float
  | CharL(l) 		-> Char
  | BoolL(l) 		-> Boolean
  | StringL(l) 	-> String


(* HEPLER FUNCTION: check variable v has type t *)
let check_type_of_expr env v t = 
  let (id, id_t) = find_var_table env.var_scope v in
  if (id_t = t) then true else false


(* HEPLER FUNCTION: check both sides of binop are semantically compatible*)
let check_compatable_types_binop t1 t2 = match (t1,t2) with
    (RCell, RInt) 				-> true
  | (RInt, RCell) 				-> true
  | (RFloat, RCell) 			-> true
  | (RCell, RFloat) 			-> true
  | (RInt, RInt) 					-> true
  | (RFloat, RFloat) 			-> true
  | (RBoolean, RBoolean) 	-> true
  | (RCell, RCell) 				-> true
  | _ -> false


(* HEPLER FUNCTION: check both sides of assign are semantically compatible*)
let check_compatable_types_assign t1 t2 = match (t1,t2) with
    (RCell, RInt) 	-> true
  | (RInt, RCell) 	-> true
  | (RFloat, RCell) -> true
  | (RCell, RFloat) -> true
  | (RLine, RInt) 	-> true
  | (RLine, RFloat) -> true
  | (tp, fp) -> if (tp = fp) then true else false


(*		 	 	let t1 = get_type_for_expr env e1 in
      	(if not ((fst t1) = RInt) then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" )));
						
	*)

(* Get type of an expression so we can semantically check it *)
let rec get_type_for_expr env e = match e with
	| Literal(i) ->
				let t = var_taty_to_ret_taty (get_type_for_literal i) in
    		(t,t)
	| TableAsTable(s,e1,e2,e3,e4) ->
				(* Check table is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 and
						t3 = get_type_for_expr env e3 and
						t4 = get_type_for_expr env e4	in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) ||
					  not ((fst t3) = RInt) ||
						not ((fst t4) = RInt)	then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" )));
    		let res = try
        		check_type_of_expr env s RTable
      	with Not_found ->
	    			raise (Error("Undeclared Table " ^ s)) in
        		(if not res then
            raise (Error("Non-Table indexed as table " ^ s)));
        let (_, und_type) = find_und_table env.und_scope s in
        (RTable, und_type)
	| TableAsLine_R(s,e1,e2,e3) -> 
				(* Check table is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 and
						t3 = get_type_for_expr env e3 in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) ||
					  not ((fst t3) = RInt) then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" )));
        let res = try
         		check_type_of_expr env s RTable
        with Not_found ->
	    			raise (Error("Undeclared Table " ^ s)) in
          	(if not res then
          	raise (Error("Non-Table indexed as table " ^ s)));
        let (_, und_type) = find_und_table env.und_scope s in
        (RLine, und_type)
	| TableAsLine_L(s,e1,e2,e3) -> 
				(* Check table is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 and
						t3 = get_type_for_expr env e3 in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) ||
					  not ((fst t3) = RInt) then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" )));
        let res = try
         		check_type_of_expr env s RTable
        with Not_found ->
	    			raise (Error("Undeclared Table " ^ s)) in
          	(if not res then
          	raise (Error("Non-Table indexed as table " ^ s)));
        let (_, und_type) = find_und_table env.und_scope s in
        (RLine, und_type)
	| TableAsCell(s,e1,e2) ->
				(* Check table is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" ))); 
         let res = try
            check_type_of_expr env s RTable
         with Not_found ->
	    			raise (Error("Undeclared Table " ^ s)) in
          	(if not res then
            raise (Error("Non-Table indexed as table " ^ s)));
         let (_, c_type) = find_und_table env.und_scope s in
         (RCell, c_type)
    | TableAsCellValue(s,e1,e2) ->
				(* Check table is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) then
      			raise (Error("Table '" ^ s ^ "' was indexed without Integers" ))); 
         let res = try
            check_type_of_expr env s RTable
         with Not_found ->
	    			raise (Error("Undeclared Table " ^ s)) in
          	(if not res then
            raise (Error("Non-Table indexed as table " ^ s)));
         let (_, c_type) = find_und_table env.und_scope s in
         (RCell, c_type)
	| LineAsLine(s,e1,e2) ->
				(* Check line is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 and 
					  t2 = get_type_for_expr env e2 in
      	(if not ((fst t1) = RInt) ||
						not ((fst t2) = RInt) then
      			raise (Error("Line '" ^ s ^ "' was indexed without Integers" ))); 
        let res = try
            check_type_of_expr env s RLine
        with Not_found ->
	    			raise (Error("Undeclared Line " ^ s)) in
          	(if not res then
            raise (Error("Non-Line indexed as line " ^ s)));
        let (_, c_type) = find_und_table env.und_scope s in
        (RLine, c_type)
	| LineAsCell(s,e1) ->
				(* Check line is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 in
      	(if not ((fst t1) = RInt) then
      			raise (Error("Line '" ^ s ^ "' was indexed without Integers" ))); 
        let res = try
        		check_type_of_expr env s RLine
        with Not_found ->
	    			raise (Error("Undeclared Line " ^ s)) in
          	(if not res then
            raise (Error("Non-Line indexed as line " ^ s)));
        let (_, c_type) = find_und_table env.und_scope s in
        (RCell, c_type)
    | LineAsCellValue(s,e1) ->
				(* Check line is indexed with integers*)
		 	 	let t1 = get_type_for_expr env e1 in
      	(if not ((fst t1) = RInt) then
      			raise (Error("Line '" ^ s ^ "' was indexed without Integers" ))); 
        let res = try
        		check_type_of_expr env s RLine
        with Not_found ->
	    			raise (Error("Undeclared Line " ^ s)) in
          	(if not res then
            raise (Error("Non-Line indexed as line " ^ s)));
        let (_, c_type) = find_und_table env.und_scope s in
        (RCell, c_type)
        
	| CellAsCell(s) ->
         let res = try
         		check_type_of_expr env s RCell
         with Not_found ->
         		raise (Error("Undeclared Line " ^ s)) in
         		(if not res then
            raise (Error("Non-Line indexed as line " ^ s)));
         let (_, c_type) = find_und_table env.und_scope s in
         (RCell, c_type) 	
         
    | NewTotalCell(s) ->
         let res = try
         		(check_type_of_expr env s RCell) || (check_type_of_expr env s RLine) || (check_type_of_expr env s RTable)
         with Not_found ->
         		raise (Error("Undeclared Table/Line/Cell " ^ s)) in
         		(if not res then
            raise (Error("^id need id to be table/line/cell " ^ s)));
         let (_, c_type) = find_und_table env.und_scope s in
         (RCell, c_type)
         
	| Id(s) -> 
          let (_, s_type) = try
              find_var_table env.var_scope s (* locate a variable by name *)
	    		with Not_found ->
	    raise (Error("Undeclared 1 Identifier " ^ s))
          in let (_, u_type) = try
              find_und_table env.und_scope s (* locate a variable by name *)
	    		with Not_found ->
	    raise (Error("Undeclared 2 Identifier " ^ s))
          in (s_type, u_type)
	| Binop(e1,op,e2) ->
          let t1 = get_type_for_expr env e1
          and t2 = get_type_for_expr env e2 in
          (if not (check_compatable_types_binop (fst t1) (fst t2)) then
              raise (Error("Mismatch in underlying types for binary operator: " ^ taty_ret_to_string (fst t1) ^ ":" ^ taty_ret_to_string (fst t2))));
          (if ((op = And || op = Or) && not ((fst t1) = RBoolean)) then
              raise (Error("Non Boolean argument passed to And / Or")));
          if op = Equal || op = Neq || op = Less_than || op = Leq || op = Greater_than || op = Geq then
              (RBoolean, get_better_und_type (snd t1) (snd t2))
          else 
              t1
	| Assign(e1,e2) ->
          let t1 = get_type_for_expr env e1
          and t2 = get_type_for_expr env e2 in
          (if not (check_compatable_types_assign (fst t1) (fst t2)) then
              raise (Error("Mismatch in types for assignment")));
       (*   (if not (check_compatable_types_assign (snd t1) (snd t2)) then
              raise (Error("Mismatch in underlying types for assignment"))); *)
          (fst t1, get_better_und_type (snd t1) (snd t2))
	| Call(s,el) ->
          let (fname, fret, fargs, fbody) = try
              find_function env.fun_scope s
          with Not_found ->
              raise (Error("Undeclared Function " ^ s)) in
          let el_ty = List.map (fun s -> fst (get_type_for_expr env s)) el in
          let fn_ty = List.map (fun s -> var_taty_to_ret_taty s) fargs in
          (* TODO - allow Cells to be passed as int / float *)
          (* let () = List.iter (print_string) (List.map (taty_ret_to_string) el_ty) in
          let () = List.iter (print_string) (List.map (taty_ret_to_string) fn_ty) in *)
          (if not (el_ty = fn_ty) then
              raise (Error("Mismatching types in function call for: " ^ s)));
          (fret, default_und_type fret)
	| Uniop(op,e1) ->
          let t1 = get_type_for_expr env e1 in
          (if not ((fst t1) = RBoolean) then
              raise (Error("Unary op applied on non-boolean")));
           t1
	| All -> (RInt, RInt)
	| Null -> (RInt, RInt)
	| UnaryMinus(e1) ->
          let ut = get_type_for_expr env e1 in
          (if not ((fst ut) = RInt || (fst ut) = RFloat || (fst ut) = RBoolean || (fst ut) = RCell) then
              raise (Error("Unary Minus used on non Int / Float / Boolean")));
          ut

let rec get_sexpr_for_expr env e = 
  let (t1,t2) = get_type_for_expr env e in
  ExprType(e,t1,t2)


(* THREE HELPER FUNCTIONS: Gets type/name/underlying type *)
(* from a variable's signature *)
let get_types_from_varSig vs =
  let VarSig(ty,nm) = vs in
  ty

let get_name_undtype_from_varSig vs =
  let VarSig(ty,nm) = vs in
  (nm, (default_und_type (var_taty_to_ret_taty ty)))

let get_name_type_from_varSig vs =
  let VarSig(ty,nm) = vs in
  (nm, (var_taty_to_ret_taty ty))


(* HELPER FUNCTION: Add function to env and return updated env *)
let add_func_to_env env func_inst = 
  let f_table = env.fun_scope in
  let old_functions = f_table.functions in
  let FuncSig(rt, nm) = func_inst.fname in
  let al = List.map (fun vs -> get_types_from_varSig vs) func_inst.formals in
  let fun_stmt = func_inst.body in
  (* let () = print_string "Adding Function: \n" in
  let () = print_string (nm ^ "\n") in
  let () = List.iter (print_string) (List.map (fun s -> (taty_ret_to_string s) ^ "\n") (List.map (var_taty_to_ret_taty) al)) in *)
  let new_functions = (nm, rt, al, fun_stmt)::old_functions in
  let new_fun_scope = {functions = new_functions} in
  let new_env = {env with fun_scope = new_fun_scope} in
  new_env


(* HELPER FUNCTION: Add a newly initialized var to env and return updated env *)
let add_init_to_env env init_inst = (* TODO - check that the variable being assigned to matches the type of thing being assigned to it *)
  (* Note that if we are initializing a Cell, then it can be initialized from either a float or int or cell *)
  let env = match init_inst with
    | NullInit(t,s) -> add_to_var_table env s (var_taty_to_ret_taty t)
    | VarInit(t,s,e) -> add_to_var_table env s (var_taty_to_ret_taty t)
    | TableInit(t1,s,e1,e2,t2) -> add_to_var_table env s (var_taty_to_ret_taty t1) in 
  let (tp,env) = match init_inst with
    | VarInit(t,s,e) -> 
      let env = add_to_und_table env s (snd (get_type_for_expr env e)) in
      (snd (get_type_for_expr env e), env)
    | TableInit(t1,s,e1,e2,t2) -> 
      let env = add_to_und_table env s (var_taty_to_ret_taty t2) in
      ((var_taty_to_ret_taty t2), env)
    | NullInit(t,s) ->
      let env = add_to_und_table env s (default_und_type (var_taty_to_ret_taty t)) in
      ((var_taty_to_ret_taty t), env) in
  (tp, env)


(* HELPER FUNCTION: Combine two environments*)
let rec combine_env env lenv =
  let ret_seen = List.fold_left (fun a e -> a && e.return_seen) true lenv in
  let new_env = {env with return_seen = ret_seen} in
  new_env
	
(* Get type of a statement so we can semantically check it *)
let rec get_env_for_stmt env st = match st with
	| Block(sl) ->
			(* If we enter a block, set break to 1, and iterate variables *)
      let new_var_scope = {parent = Some(env.var_scope); variables = [] } in
      let new_env = {env with var_scope = new_var_scope} in
      let getter (env, acc) s =
        let (st, ne) = get_env_for_stmt env s in
        (ne, st::acc) in
      let (ls, st) = List.fold_left (fun e s -> getter e s) (new_env,[]) sl in      
      let revst = List.rev st in
      (SBlock(revst), ls)
	| Expr(e) ->
      let _ = get_type_for_expr env e in
      (SExpr(get_sexpr_for_expr env e), env)
	| Return(e) ->
      let t = get_type_for_expr env e in
      (if not ((fst t) = env.return_type) then
      		raise (Error("Incompatable Return type")));
      let new_env = {env with return_seen = true} in
      (SReturn(get_sexpr_for_expr env e), new_env)
	| Print(e) ->
      let _ = get_type_for_expr env e in
      (* let () = print_string "Print Called \n\n" in *)
      (* (if not (t = RTable) then
      raise (Error("Incompatable Print type"))); *)
      (SPrint(get_sexpr_for_expr env e), env)
	| If(e,s1,s2) ->
      let t = get_type_for_expr env e in
      (if not ((fst t) = RBoolean) then
      		raise (Error("Non-boolean IF predicate")));
      let new_env = {env with break_possible = true} in
      let (st1, new_env1) = get_env_for_stmt new_env s1
      and (st2, new_env2) = get_env_for_stmt new_env s2 in
      let ret_seen = (new_env1.return_seen && new_env2.return_seen) in
      let new_env = {env with return_seen = ret_seen} in
      (SIf((get_sexpr_for_expr env e), st1, st2), new_env)
	| For(e1,e2,e3,s) ->
      let t1 = fst (get_type_for_expr env e1)
      and t2 = fst (get_type_for_expr env e2)
      and t3 = fst (get_type_for_expr env e3) in
      (if not (t1 = RInt && t3 = RInt && t2 = RBoolean) then
      		raise (Error("Bad FOR loop, not integer bounds")));
      (if not (t1 = t3) then
      		raise (Error("Bad FOR loop, can not assign to iterator")));
      let new_env = {env with break_possible = true} in
      let (st, new_env) = get_env_for_stmt new_env s in
      (SFor((get_sexpr_for_expr env e1), (get_sexpr_for_expr env e2), (get_sexpr_for_expr env e3), st), new_env)
	| While(e,s) ->
      let t = fst (get_type_for_expr env e) in
      (if not (t = RBoolean) then
      		raise (Error("Bad WHILE loop, not integer bounds")));
      let new_env = {env with break_possible = true} in
      let (st, new_env) = get_env_for_stmt new_env s in
      (SWhile((get_sexpr_for_expr env e), st), new_env)
	| NoElseMark -> (SNoElseMark, env)
	| VarDecl(i) ->
      let (si, new_env) = add_init_to_env env i in
      (SVarDecl(SInit(i,si)), new_env)
	| Break ->
      (if not (env.break_possible) then
      		raise (Error("Bad BREAK statement")));
      (SBreak, env)


(* Add all global vars and functions to our environment *)
let initialize_globals env constr = match constr with
    GlobalVar(i) ->
      let (_, new_env) = add_init_to_env env i in
      new_env
  | FuncDef(f) ->
      let new_env = add_func_to_env env f in
      new_env


(* Check that return statements are in order, etc. *)
let check_final_env env = (* TODO - do we need to do anything else here? *)
  (if (false = env.return_seen && env.return_type <> RVoid) then
    raise (Error("Missing Return Statement")));
  true


(* Semantic checking on a function *)
let check_funcs env ct = match ct with
    GlobalVar(v) ->
      let (tp, _) = add_init_to_env env v in
      SGlobalVar(SInit(v, tp))
  | FuncDef(f) ->
      let FuncSig(ret_type,fun_name) = f.fname in
      (* let () = print_string ("Checking function: " ^ fun_name ^ "\n\n") in *)
      let new_variables = List.fold_left (fun a vs -> (get_name_type_from_varSig vs)::a) [] f.formals in
      let new_und_variables = List.fold_left (fun a vs -> (get_name_undtype_from_varSig vs)::a) [] f.formals in
      let new_var_scope = {parent = Some(env.var_scope); variables = new_variables} in
      let new_und_scope = {parent = Some(env.und_scope); variables = new_und_variables} in 
      let new_env = {return_type = ret_type; return_seen = false; var_scope = new_var_scope; und_scope = new_und_scope; fun_scope = env.fun_scope; break_possible = false} in
      let (stbody, final_env) = get_env_for_stmt new_env f.body in
      let _ = check_final_env final_env in
      let sfunc = {sfname = f.fname; sformals = f.formals; sbody = stbody} in
      SFuncDef(sfunc)



(* Empty initializations for tables/environments *)
let empty_symbol_table = {parent = None; variables = []}
let empty_function_table = {functions = []}
let empty_env = {return_type = RVoid; return_seen = false; var_scope = empty_symbol_table; und_scope = empty_symbol_table; fun_scope = empty_function_table; break_possible = false}


(* MAIN RUNNER FUNCTION *)
let translate prog =
  let env = List.fold_left (fun env ct -> initialize_globals env ct) empty_env prog in
  let res = List.map (fun ct -> check_funcs env ct) prog in
  res
