//////////////////////////////////////////////////////////////////////////////////
// 	Table Type - 2d array composed of Cells	  									// 
// NOTE: if the table is holding floats, all numbers must end with				//
// an "f".  For example, myTable.setCellValue(1, 2, 3.5f)						//
// This is because Java automatically assumes decimal point numbers				//
// are double unless explicitly case to float -- resulting in compile errors	//
//////////////////////////////////////////////////////////////////////////////////


package java_lib;

// E must be Integer or Float
public class Table<E> {
	private Cell<E>[][] table;
	private int numRows;
	private int numColumns;

	
	//////////////////////////////////////////////////////////////////////////////////
	//	Example TaML to Java conversions:											//
	//	table t1 = ([3,4], int); --> Table<Integer> t1 = new Table<Integer>(3,4);	//
	//	table t2 = ([3,4], float); --> Table<Float> t2 = new Table<Float>(3,4);		//
	//////////////////////////////////////////////////////////////////////////////////
	public Table(int numR, int numC){
		numRows = numR;
		numColumns = numC;
		
		// Error Checking
		if(numRows>500 || numColumns>50)
			throw new IllegalArgumentException("Error: There can't be more than 500 rows or 50 columns");
		if(numRows<1 || numColumns<1)
			throw new IllegalArgumentException("Error: You must have at least 1 column and/or row");

		// Fill the table with Cell objects
		table = new Cell[numRows][numColumns];
		for (int i=0; i<numRows; i++){
			for(int j=0; j<numColumns; j++){
				Cell<E> c = new Cell<E>();
				table[i][j] = c;
			}
		}
	}
	
	
	//======================================================================//
	//				TABLE COPIERS											//
	// Creates & returns a new table that is a COPY of cells in 			//
	// the old table. (e.g., table has new memory allocation)				//
	// Note that the new Table will be of the same type (int or float)  	//
	// as the old table.													//
	//======================================================================//
	
	
	//////////////////////////////////////////////////////////////////////	
	// Example Usage: Table newT = oldT.createTableCopy(1, 5, 1, 4);	//
	//////////////////////////////////////////////////////////////////////
	public Table<E> createTableCopy(int startRow, int endRow, int startCol, int endCol){
		//Error Checking
		if(endRow<startRow || endCol<startCol )
			throw new IllegalArgumentException("Error start row/column must be less than end row/column");	
		
		// Initialize the new table
		int newNumRows = (endRow - startRow) + 1;
		int newNumCols = (endCol - startCol) + 1;
		Table<E> newTable = new Table<E>(newNumRows,newNumCols);
		
		// Copy cells into the new table
		copyCells(startRow, endRow, startCol, endCol, newTable);
		return newTable;
	}
	
	
	//////////////////////////////////////////////////////////////////////	
	// Example Usage: Table newT = oldT.createTableCopy(ALL, ALL, 1, 4);//
	//////////////////////////////////////////////////////////////////////
	public Table<E> createTableCopy(String a1, String a2, int startCol, int endCol){
		//Error Checking
		if(endCol < startCol)
			throw new IllegalArgumentException("Error start column must be less than end column");	
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both rows are not @");	
		
		// Initialize the new table
		int newNumRows = numRows;
		int newNumCols = (endCol - startCol) + 1;
		Table<E> newTable = new Table<E>(newNumRows,newNumCols);
		
		// Copy cells into the new table
		copyCells(0, newNumRows-1, startCol, endCol, newTable);
		return newTable;
	}
	
	
	//////////////////////////////////////////////////////////////////////	
	// Example Usage: Table newT = oldT.createTableCopy(1, 4, ALL, ALL) //
	//////////////////////////////////////////////////////////////////////
	public Table<E> createTableCopy(int startRow, int endRow, String a1, String a2){
		//Error Checking
		if(endRow < startRow)
			throw new IllegalArgumentException("Error start row must be less than end row");	
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both columns are not @");	
		
		// Initialize the new table
		int newNumRows = (endRow - startRow) + 1;
		int newNumCols = numColumns;
		Table<E> newTable = new Table<E>(newNumRows,newNumCols);
		
		// Copy cells into the new table
		copyCells(startRow, endRow, 0, newNumCols-1, newTable);
		return newTable;
	}
	
	//////////////////////////////////////////////////////////////////////	
	// Example: Table newT = oldT.createTableCopy(ALL, ALL, ALL, ALL)  //
	//////////////////////////////////////////////////////////////////////
	public Table<E> createTableCopy(String a1, String a2, String a3, String a4){
		//Error Checking
		if(!a1.equals("ALL")||!a2.equals("ALL")||!a3.equals("ALL")||!a4.equals("ALL"))
			throw new IllegalArgumentException("Error -- both rows/columns are not @");	
		
		// Initialize the new table
		Table<E> newTable = new Table<E>(numRows,numColumns);
		
		// Copy cells into the new table
		copyCells(0, numRows-1, 0, numColumns-1, newTable);
		return newTable;
	}
	
	
	// HELPER FUNCTION: for copying cells into new table
	private void copyCells(int startRow, int endRow, int startCol, int endCol,Table<E> newTable) {
		int a=0;				// newTable's row index
		int b=0;				// newTable's column index
		
		for (int i=startRow; i<=endRow; i++){
			b = 0;
			for(int j=startCol; j<=endCol; j++){
				Cell<E> c = new Cell<E>();
				c.setVal(table[i][j].getVal());
				newTable.setCell(a, b, c);
				b++;
			}
			a++;
		}
	}
	
	
			
	//======================================================================//
	//				GETTERS & SETTERS						   				//
	//======================================================================//	
	public E getCellValue(int row, int column){
		return table[row][column].getVal();
	}
	
	// Set single cell in table to a certain value
	public void setVal(int row, int column, E newValue){
		table[row][column].setVal(newValue);
	}
	
	// Sets entire table to a single value							
	public void setVal( E newValue){
		for (int i=0; i<numRows; i++){
			for(int j=0; j<numColumns; j++)
				table[i][j].setVal(newValue);
		}
	}
	
	// Sets subrange of table to a single value. Index ranges ex: [1~2, 3~5]
	public void setVal(int startRow, int endRow, int startCol, int endCol, E newValue) {
		if(endRow<startRow || endCol<startCol )
			throw new IllegalArgumentException("Error start row/column must be less than end row/column");	
		
		for (int i=startRow; i<=endRow; i++){
			for(int j=startCol; j<=endCol; j++)
				table[i][j].setVal(newValue);
		}
	}
	
	// Sets subrange of table to a single value. Index ranges ex: [@, 3~5]
	public void setVal(String a1, String a2, int startCol, int endCol, E newValue) {
		if(endCol < startCol)
			throw new IllegalArgumentException("Error start column must be less than end column");	
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both rows are not @");		
		
		for (int i=0; i<numRows; i++){
			for(int j=startCol; j<=endCol; j++)
				table[i][j].setVal(newValue);
		}
	}
	
	// Sets subrange of table to a single value. Index ranges ex: [3~5, @]
	public void setVal(int startRow, int endRow, String a1, String a2, E newValue) {
		if(endRow < startRow)
			throw new IllegalArgumentException("Error start row must be less than end row");		
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both rows are not @");		
		
		for (int i=startRow; i<=endRow; i++){
			for(int j=0; j<numColumns; j++)
				table[i][j].setVal(newValue);
		}
	}
	
	
	
	//////////////////////////////////////////////////////////////
	// Useful if we want to set a cell/line segment to a cell	//
	// in the table.  e.g., "cell c = myTable[3,2]" could		//
	// leverage this function									//
	//////////////////////////////////////////////////////////////
	public Cell<E> getCell(int row, int column){
		return table[row][column];
	}
	
	// Private since this is only used internally when copying
	// cells into a new table.  Otherwise, users really shouldn't
	// be able to copy cells into tables -- they should only be
	// allowed to edit the values held by those cells (via 
	// setCellValue etc. )
	private void setCell(int row, int column, Cell<E> c){
		table[row][column] = c;
	}
	
	
	public int numRows(){
		return numRows;
	}
	
	public int numColumns(){
		return numColumns;
	}
	
	
	//========================================================================//
	//				PRINT FUNCTION							   				  //
	//========================================================================//	
	public void print(){
		StringBuffer[] rowText = new StringBuffer[numRows+3]; 	// text of each Row
		int maxLength;											// max length of a particular column
		String bars = "===============================================";
		
		// Setup Row Values & initialize StringBuffers
		rowText[0] = new StringBuffer("" + '\t');
		rowText[1] = new StringBuffer("" + '\t');
		for (int i=2; i<=numRows+1; i++)
			rowText[i] = new StringBuffer(String.valueOf(i-1) + '\t' + "| ");
		rowText[numRows+2] = new StringBuffer("" + '\t');
		
		// Save Data for each column into rowText[]
		for (int j=0; j<numColumns; j++){
			
			// Find Max Length in a certain column
			maxLength = 0;
			for (int i=0; i<numRows; i++){
				int curLength = String.valueOf(table[i][j].getVal()).length();
				if(curLength > maxLength)
					maxLength = curLength;
			}
			
			// Save Data in each row, setting length of column to fit maxLength
			for (int i=0; i<numRows; i++){
				String curNum = String.valueOf(table[i][j].getVal());
				String paddedNum = String.format("%1$-" + maxLength + "s", curNum);
				rowText[i+2].append(paddedNum + " | ");
			}
			
			// Save Data for Column Labels
			String curLabel;
			if(j<26)
				curLabel = String.valueOf((char)(j + 65));
			else
				curLabel = String.valueOf((char)(j + 39))+String.valueOf((char)(j + 39));
			String paddedLabel = String.format("%1$-" + maxLength + "s", curLabel);
			rowText[0].append("  " + paddedLabel + " ");			
			
			// Save Data for Bottom & Top Bars
			String bottomBar = bars.substring(0, maxLength+3); 
			rowText[1].append(bottomBar);
			rowText[numRows+2].append(bottomBar);
		}
		
		// Print rows
		for (int i=0; i<=numRows+2; i++){
			System.out.println(rowText[i]);
		}	
	}
		
	
	
}
